import EventEmitter from 'events';
import * as fs from 'fs';
import { exec } from 'child_process';
import { gt } from 'semver';

function sequence ( array, fn ) {
	var results = [];
	var promise = Promise.resolve();

	function next ( member, i ) {
		return fn( member ).then( function (value) { return results[i] = value; } );
	}

	var loop = function ( i ) {
		promise = promise.then( function () { return next( array[i], i ); } );
	};

	for ( var i = 0; i < array.length; i += 1 ) loop( i );

	return promise.then( function () { return results; } );
}

var name = "rollup-watch";
var version = "2.2.0";

function checkVersion ( name, localVersion ) {
	return new Promise( function ( fulfil, reject ) {
		exec( ("npm show " + name + " version"), function ( err, result ) {
			if ( err ) return reject( err );
			fulfil( result.trim() );
		});
	}).then( function (latestVersion) {
		if ( gt( latestVersion, localVersion ) ) {
			var err = new Error( (name + " is out of date") );
			err.code = 'OUT_OF_DATE';
			err.localVersion = localVersion;
			err.latestVersion = latestVersion;
		}
	});
}

var FileWatcher = function FileWatcher ( file, data, callback ) {
	fs.watch( file, { encoding: 'utf-8', persistent: true }, function () {
		// this is necessary because we get duplicate events...
		var contents = fs.readFileSync( file, 'utf-8' );
		if ( contents !== data ) {
			data = contents;
			callback();
		}
	});
};

function watch ( rollup, options ) {
	var emitter = new EventEmitter();

	process.nextTick( function () { return emitter.emit( 'event', { code: 'STARTING' }); } );

	checkVersion( name, version )
		.catch( function (err) {
			if ( err.code === 'OUT_OF_DATE' ) {
				// TODO offer to update
				console.error( ("rollup-watch is out of date (you have " + (err.localVersion) + ", latest version is " + (err.latestVersion) + "). Update it with npm install -g rollup-watch") ); // eslint-disable-line no-console
			}
		})
		.then( function () {
			var filewatchers = new Map();

			var rebuildScheduled = false;
			var building = false;
			var watching = false;

			var timeout;

			function triggerRebuild () {
				clearTimeout( timeout );
				rebuildScheduled = true;

				timeout = setTimeout( function () {
					if ( !building ) {
						rebuildScheduled = false;
						build();
					}
				}, 50 );
			}

			function build () {
				if ( building ) return;

				var start = Date.now();
				var initial = !watching;

				emitter.emit( 'event', { code: 'BUILD_START' });

				building = true;

				return rollup.rollup( options )
					.then( function (bundle) {
						bundle.modules.forEach( function (module) {
							var id = module.id;

							if ( !filewatchers.has( id ) ) {
								var watcher = new FileWatcher( id, module.originalCode, triggerRebuild );
								filewatchers.set( id, watcher );
							}
						});

						if ( options.targets ) {
							return sequence( options.targets, function (target) {
								var mergedOptions = Object.assign( {}, options, target );
								return bundle.write( mergedOptions );
							});
						}

						return bundle.write( options );
					})
					.then( function () {
						emitter.emit( 'event', {
							code: 'BUILD_END',
							duration: Date.now() - start,
							initial: initial
						});
					}, function (error) {
						emitter.emit( 'event', {
							code: 'ERROR',
							error: error
						});
					})
					.then( function () {
						building = false;
						if ( rebuildScheduled ) build();
					});
			}

			build();
		});

	return emitter;
}

export default watch;